import functools
import webbrowser
import time
import json
import sys
from typing import Dict, List
import logging
import uuid
from pathlib import Path
import os
from suds import WebFault

import backoff
from bingads.v13.bulk import (
    AuthorizationData,
    FileDownloadException,
    SdkException,
    OAuthTokenRequestException,
    OAuthDesktopMobileAuthCodeGrant
)
from bingads.v13.reporting import ServiceClient, ReportingServiceManager
from arcane.core.exceptions import _exceptions, GOOGLE_EXCEPTIONS_TO_RETRY

from .helpers import parse_webfault_errors, parse_bing_response
from .exceptions import MicrosoftAdvertisingAccountLostAccessException, MICROSOFT_EXCEPTIONS_TO_RETRY


TEMP_FILE_DIRECTORY = "/tmp/"
WORKING_DIRECTORY = "/tmp/BingAdsSDKPython"
BING_API_VERSION = 13


def authenticate(func):
    @functools.wraps(func)
    def _authenticate(self, *args, **kwargs):
        # You should authenticate for Bing Ads services with a Microsoft Account,
        # instead of providing the Bing Ads username and password set.
        try:
            self.authenticate_with_oauth()
        except EOFError:
            time.sleep(5)
            self.authenticate_with_oauth()

        if not os.path.exists(WORKING_DIRECTORY):
            try:
                os.makedirs(WORKING_DIRECTORY)
            except FileExistsError:
                # In case of several threads that are trying to create the tmp bing directory simultaneously.
                pass

        return func(self, *args, **kwargs)
    return _authenticate


class Client:
    def __init__(
        self,
        credentials,
        secrets_bucket,
        refresh_token_location,
        storage_client,
        customer_id=None,
        account_id=None
    ):
        super()
        if credentials is not None:
            self.credentials = json.load(open(credentials))
            self.authorization_data = AuthorizationData(
                account_id=account_id,
                customer_id=customer_id,
                developer_token=self.credentials["developer_token"]
            )
        else:
            logging.warning('WARNING: Empty credentials provided')
        self.secrets_bucket = secrets_bucket
        self.refresh_token_location = refresh_token_location
        self.storage_client = storage_client

    @authenticate
    @backoff.on_exception(backoff.expo, (MICROSOFT_EXCEPTIONS_TO_RETRY), max_tries=3)
    def get_bing_ads_api_client(self):

        reporting_service_manager = ReportingServiceManager(
            authorization_data=self.authorization_data,
            poll_interval_in_milliseconds=5000,
            environment=self.credentials["environment"],
            working_directory=WORKING_DIRECTORY
        )

        # In addition to ReportingServiceManager, you will need a reporting ServiceClient
        # to build the ReportRequest.

        reporting_service = ServiceClient(
            'ReportingService',
            authorization_data=self.authorization_data,
            environment=self.credentials["environment"],
            version=13,
        )

        return reporting_service_manager, reporting_service

    @authenticate
    @backoff.on_exception(backoff.expo, (MICROSOFT_EXCEPTIONS_TO_RETRY), max_tries=3)
    def get_service_client(self, service_name: str, api_version: int = BING_API_VERSION):
        return ServiceClient(
            service_name,
            authorization_data=self.authorization_data,
            environment=self.credentials["environment"],
            version=api_version,
        )

    @backoff.on_exception(backoff.expo, (MICROSOFT_EXCEPTIONS_TO_RETRY), max_tries=3)
    def get_bing_campaigns(self, bing_account_id: int) -> List[Dict]:
        """ Get account campaigns (id, name and status) """
        campaign_service = self.get_service_client(
            service_name='CampaignManagement')

        campaign_types = ['Search Shopping DynamicSearchAds Audience']
        try:
            response = campaign_service.GetCampaignsByAccountId(
                AccountId=bing_account_id,
                CampaignType=campaign_types
            )
        except WebFault as e:
            print(str(parse_webfault_errors(e)))
            raise MicrosoftAdvertisingAccountLostAccessException(
                f'Could not access account {bing_account_id}. Are you sure you entered the correct id?'
            )

        all_campaigns = [
            {
                "id": campaign['Id'],
                "name": campaign['Name'],
                "status": campaign['Status']
            }
            for campaign in parse_bing_response(response).get('Campaign', [])
        ]

        return all_campaigns

    @backoff.on_exception(backoff.expo, (MICROSOFT_EXCEPTIONS_TO_RETRY), max_tries=3)
    def get_account_name(self, bing_account_id: int) -> str:
        """Get account name. This is usefull to check account access."""

        customer_service = self.get_service_client(
            service_name='CustomerManagement')

        try:
            response = customer_service.GetAccount(AccountId=bing_account_id)
        except WebFault as e:
            print(str(parse_webfault_errors(e)))
            raise MicrosoftAdvertisingAccountLostAccessException(
                f'Could not access account {bing_account_id}. Are you sure you entered the correct id?'
            )

        return response.Name


    def authenticate_with_oauth(self):
        authentication = OAuthDesktopMobileAuthCodeGrant(
            client_id=self.credentials["client_id"],
            env=self.credentials["environment"]
        )

        # It is recommended that you specify a non guessable 'state' request parameter to help prevent
        # cross site request forgery (CSRF).
        authentication.state = 'ClientStateGoesHere'

        # Assign this authentication instance to the self.authorization_data.
        self.authorization_data.authentication = authentication

        # Register the callback function to automatically save the refresh token anytime it is refreshed.
        # Uncomment this line if you want to store your refresh token. Be sure to save your refresh token securely.
        self.authorization_data.authentication.token_refreshed_callback = self.save_refresh_token

        refresh_token = self.get_refresh_token()

        try:
            # If we have a refresh token let's refresh it
            if refresh_token is not None:
                self.authorization_data.authentication.request_oauth_tokens_by_refresh_token(refresh_token)
            else:
                self.request_user_consent()
        except OAuthTokenRequestException:
            # The user could not be authenticated or the grant is expired.
            # The user must first sign in and if needed grant the client application access to the requested scope.
            self.request_user_consent()

    def request_user_consent(self):
        webbrowser.open(self.authorization_data.authentication.get_authorization_endpoint(), new=1)
        if sys.version_info.major >= 3:
            response_uri = input(
                "You need to provide consent for the application to access your Bing Ads accounts. "
                "After you have granted consent in the web browser for the application to access your Bing Ads accounts, "
                "please enter the response URI that includes the authorization 'code' parameter: \n"
            )
        if self.authorization_data.authentication.state != self.credentials["client_state"]:
            raise Exception("The OAuth response state does not match the client request state.")

        # Request access and refresh tokens using the URI that you provided manually during program execution.
        self.authorization_data.authentication.request_oauth_tokens_by_response_uri(response_uri=response_uri)

    @backoff.on_exception(backoff.expo, (GOOGLE_EXCEPTIONS_TO_RETRY), max_tries=3)
    def get_refresh_token(self):
        blob_token = self.storage_client.bucket(self.secrets_bucket).blob(self.refresh_token_location)
        try:
            return blob_token.download_as_string()
        except _exceptions.NotFound:
            return None

    @backoff.on_exception(backoff.expo, (GOOGLE_EXCEPTIONS_TO_RETRY), max_tries=3)
    def save_refresh_token(self, oauth_tokens):
        """ Stores a refresh token on a bucket. """
        blob_token = self.storage_client.bucket(self.secrets_bucket).blob(self.refresh_token_location)
        blob_token.upload_from_string(oauth_tokens.refresh_token)

    @backoff.on_exception(backoff.expo, (*MICROSOFT_EXCEPTIONS_TO_RETRY, FileDownloadException, SdkException), max_tries=3)
    def submit_and_download(self, report_request, reporting_service_manager: ReportingServiceManager) -> str:
        """
        Submit the download request and then store the result file in a local temp file.
        Returns the path to the temporary downloaded file
        :param report_request: created with reporting_service.factory.create('<...ReportRequest>')
        :raises: FileDownloadException if we failed to download the file in a reasonable amount of time
        """

        temp_dir = Path(TEMP_FILE_DIRECTORY)
        if not temp_dir.is_dir():
            os.mkdir(TEMP_FILE_DIRECTORY)

        reporting_download_operation = reporting_service_manager.submit_download(report_request)
        # You may optionally cancel the track() operation after a specified time interval.
        reporting_download_operation.track(timeout_in_milliseconds=3600000)
        temp_file_name = str(uuid.uuid4()) + 'temp-local-report.csv'
        max_ms_to_download_report = 10 * 60 * 1000  # 10 minutes
        result_file_path = reporting_download_operation.download_result_file(
            result_file_directory=TEMP_FILE_DIRECTORY,
            result_file_name=temp_file_name,
            decompress=True,
            overwrite=True,  # Set this value true if you want to overwrite the same file.
            timeout_in_milliseconds=max_ms_to_download_report
            # You may optionally cancel the download after a specified time interval.
        )

        print("Download temporary result file: {0}\n".format(result_file_path))

        return result_file_path



from typing import Union, List
from dataclasses import dataclass


@dataclass
class BingErrorDetail:
    code: Union[int, None] = None
    details: Union[str, None] = None
    error_code: Union[str, None] = None
    message: Union[str, None] = None


@dataclass
class BingError:
    tracking_id: str
    errors: List[BingErrorDetail]



from typing import List
from suds import WebFault
from suds.sudsobject import asdict

from .types import BingError, BingErrorDetail


def parse_bing_response(response):
    """Parse Suds object response into a dict"""
    return _recursive_asdict(response)


def _recursive_asdict(d):
    """
    Convert Suds object into serializable format.

    Courtesy of https://stackoverflow.com/a/48928384
    """
    out = {}
    for k, v in asdict(d).items():
        if hasattr(v, '__keylist__'):
            out[k] = _recursive_asdict(v)
        elif isinstance(v, list):
            out[k] = []
            for item in v:
                if hasattr(item, '__keylist__'):
                    out[k].append(_recursive_asdict(item))
                elif not isinstance(item, list):
                    out[k] = item
                else:
                    out[k].append(item)
        else:
            out[k] = v
    return out


def parse_webfault_errors(ex: WebFault) -> BingError:
    """From https://docs.microsoft.com/en-us/advertising/guides/walkthrough-desktop-application-python?view=bingads-13"""

    if not hasattr(ex.fault, "detail"):
        raise Exception("Unknown WebFault")

    error_attribute_sets = (
        ["ApiFault", "OperationErrors", "OperationError"],
        ["AdApiFaultDetail", "Errors", "AdApiError"],
        ["ApiFaultDetail", "BatchErrors", "BatchError"],
        ["ApiFaultDetail", "OperationErrors", "OperationError"],
        ["EditorialApiFaultDetail", "BatchErrors", "BatchError"],
        ["EditorialApiFaultDetail", "EditorialErrors", "EditorialError"],
        ["EditorialApiFaultDetail", "OperationErrors", "OperationError"],
    )

    bing_errors: List[BingErrorDetail] = []
    for error_attribute_set in error_attribute_sets:
        bing_errors += _parse_error_detail(ex.fault.detail, error_attribute_set)

    tracking_id = ex.document.childAtPath('Envelope/Body/Fault/detail/AdApiFaultDetail/TrackingId')
    if tracking_id is not None:
        tracking_id = tracking_id.getText()

    return BingError(
        tracking_id=tracking_id,
        errors=bing_errors
    )


def _parse_bing_ads_webfault_error(error) -> BingErrorDetail:
    bing_error = BingErrorDetail()
    if hasattr(error, 'ErrorCode'):
        bing_error.error_code = error.ErrorCode
    if hasattr(error, 'Code'):
        bing_error.code = error.Code
    if hasattr(error, 'Details'):
        bing_error.details = error.Details
    if hasattr(error, 'Message'):
        bing_error.message = error.Message
    return bing_error


def _parse_error_detail(error_detail, error_attribute_set) -> List[BingErrorDetail]:
    bing_errors: List[BingErrorDetail] = []
    api_errors = error_detail
    for field in error_attribute_set:
        api_errors = getattr(api_errors, field, None)
    if api_errors is None:
        return bing_errors
    if isinstance(api_errors, list):
        for api_error in api_errors:
            bing_errors.append(_parse_bing_ads_webfault_error(api_error))
    else:
        bing_errors.append(_parse_bing_ads_webfault_error(api_errors))
    return bing_errors



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['arcane', 'arcane.bing']

package_data = \
{'': ['*']}

install_requires = \
['arcane-core>=1.6.0,<2.0.0', 'backoff>=1.10.0,<2.0.0', 'bingads==13.0.13']

setup_kwargs = {
    'name': 'arcane-bing',
    'version': '0.4.0',
    'description': 'Helpers to request bing API',
    'long_description': "# Arcane bing\n\nThis package is based on [bingads](https://docs.microsoft.com/en-us/advertising/guides/request-download-report?view=bingads-13).\n\n## Get Started\n\n```sh\npip install arcane-bing\n```\n\n## Example Usage\n\n### Reporting\n\n```python\nbing_client = Client(\n    credentials=Config.BING_ADS_CREDENTIALS,\n    secrets_bucket=Config.SECRETS_BUCKET,\n    refresh_token_location=Config.BING_ADS_REFRESH_TOKEN,\n    storage_client=storage_client\n)\n\nreporting_service_manager, reporting_service = bing_client.get_bing_ads_api_client()\n\nreport_request = build_campaigns_report(reporting_service, bing_account_id)\n\nresult_file_path = bing_client.submit_and_download(report_request, reporting_service_manager)\n```\n\n### Campaign Service\n\n:warning: For some API methods, you must provide the client's account id and the manager's customer id\n\n```python\nfrom arcane.bing import Client\nfrom arcane.bing.helpers import parse_webfault_errors, parse_bing_response\n\n\nbing_client = Client(\n    credentials=Config.BING_ADS_CREDENTIALS,\n    secrets_bucket=Config.SECRETS_BUCKET,\n    refresh_token_location=Config.BING_ADS_REFRESH_TOKEN,\n    storage_client=storage_client,\n    customer_id=CUSTOMER_ID,\n    account_id=ACCOUNT_ID\n)\n\ncampaign_service = bing_client.get_service_client(service_name='CampaignManagement')\n\ntry:\n    response = campaign_service.GetCampaignsByAccountId(AccountId=ACCOUNT_ID)\n    all_campaigns = parse_bing_response(response)['Campaign']\n    # do stuff with all_campaigns\nexcept WebFault as e:\n    bing_error = parse_webfault_errors(e)\n    # do stuff with bing_error\n```\n",
    'author': 'Arcane',
    'author_email': 'product@arcane.run',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
